/*
 * Copyright (c) 2007 Luigi Rizzo - Universita` di Pisa
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Id: linux_compat.c,v 1.63 2007/01/30 23:20:16 luigi Exp $
 *
 * This file implement an adaptation layer to compile linux usb device
 * drivers as loadable kernel modules on FreeBSD.
 * The headers provide macros and definitions, whereas this file maps the
 * internal linux API into equivalent (more or less) FreeBSD calls. 
 *
 * compile with the following macros
 *
 * -DDEV_NAME="video"	device name
 * -DDRIVER_NAME="foo"	driver name
 * make sure ldev_devs is the array of ldev_info with vendor/prod identifiers
 * to be used for matching the device.
 */

#include <ldev_stub.h>

#include <linux/module.h>
#include <linux/usb.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>		/* vmalloc_to_page */
#include <linux/sched.h>	/* schedule */

/*
 * note- here we use the linux-like 'struct usb device'
 * not the freebsd one
 */

#include <linux/jiffies.h>	/* jiffies support */

/*
 * The first part of this file is made of routines called by the FreeBSD
 * template driver.
 */

static const char *
eptype(int t)
{
	static const char *names[] = { "CONTROL", "ISOC", "BULK", "INTR" };
	return names[t & 3];
}

/*
 * Parse the configuration. If passed a non-null LD, also allocate
 * and fill the struct needed for linux (for interface ifno.
 * This is done in two passes: in the first pass count items, then allocate
 * a single block of memory, then fill memory as required.
 * Printing is enabled on first pass if ld = NULL.
 * We only allocate for interface ifno, even if printing occurs
 * for all of them.
 */ 
static void
parse_full_config(usb_config_descriptor_t *cdesc, int ifno,
	struct linux_usb_descriptors *ld)
{
	int pass, lim = (ld ? 2 : 1);		/* how many passes */
	int nintf = 0;	/* total number of interfaces */
	/* for our interface: number of alternate configs (nalt),
	 * number of endpoints for each config (neps) and total
	 * number of endpoints (nep).
	 */
	int ifbitmap = 0;	/* bitmap of interfaces we found */
	int nalt = 0, neps[16], nep = 0, cur_alt = -1;

	/*
	 * these two are used in the second pass to update the
	 * relevant linux emulation descriptors
	 */
	struct usb_host_endpoint *next_ep = NULL;
	struct usb_host_interface *next_alt = NULL;

	if (cdesc == NULL)
		return;
	bzero(neps, sizeof(neps));

    for (pass = 0; pass < lim; pass++) {

 	char *p = (char *)cdesc;
	char *end = p + UGETW(cdesc->wTotalLength);
        usb_interface_descriptor_t *id = (usb_interface_descriptor_t *)p;
        usb_device_descriptor_t *dd;
        usb_config_descriptor_t *cd;
        usb_string_descriptor_t *sd;
        usb_endpoint_descriptor_t *ed;
	int ifnum;

	int do_print = 0 && (lim == 1);	/* XXX set 0 to 1 to print */

        for (; p + id->bLength <= end; p += id->bLength) {
                id = (usb_interface_descriptor_t *)p;
		switch(id->bDescriptorType) {
		default:
			if (do_print)	/* print only */
				printf("unknown desc 0x%x len %d ",
				    id->bDescriptorType, id->bLength);
			break;

		case UDESC_DEVICE:
			dd = (usb_device_descriptor_t *)p;
			if (do_print)
				printf("DEVICE, nconf %d proto %d maxpkt 0x%x",
					dd->bNumConfigurations,
					dd->bDeviceProtocol,
					dd->bMaxPacketSize);
			break;
				
		case UDESC_CONFIG:
			cd = (usb_config_descriptor_t *)p;
			if (do_print)
			    printf("CONFIG, int %d confval %d conf %d",
				cd->bNumInterface,
				cd->bConfigurationValue,
				cd->iConfiguration);
			break;

		case UDESC_STRING:
			sd = (usb_string_descriptor_t *)p;
			if (do_print)
			    printf("STRING, len %d",
				sd->bLength);
			break;

		case UDESC_INTERFACE:
			id = (usb_interface_descriptor_t *)p;
			ifnum = id->bInterfaceNumber;
			if (pass == 0 && (ifbitmap & (1<<ifnum)) == 0) {
				/* a new interface */
				nintf++;
				ifbitmap |= (1<<ifnum);
			}
			if (ifno == -1 && ifnum != ifno) /* not one we care */
			    cur_alt = -1;
			else {
			    cur_alt = id->bAlternateSetting;
			    if (pass == 1) {		/* set */
				/* second pass, set the endpoints */
				next_alt->desc =
					*(struct usb_interface_descriptor *)p;
				next_alt->endpoint = next_ep;
				if (0)
				    printf("alt %d has %d enpoints (%d)\n",
					cur_alt, neps[cur_alt],
					next_alt->desc.bNumEndpoints);
				next_alt++;
			    }
			    nalt++;
			}
			if (do_print) {
			    printf("INTF, num %d alt %d n_ep %d",
				id->bInterfaceNumber,
				id->bAlternateSetting,
				id->bNumEndpoints);
			}
			break;

		case UDESC_ENDPOINT:
			ed = (usb_endpoint_descriptor_t *)p;
			if (pass == 0 && cur_alt != -1) {
				nep++;
				neps[cur_alt]++;
			}
			if (pass == 1) { /* fill the endpoint with *ed */
				next_ep->desc =
					*(struct usb_endpoint_descriptor *)p;
				next_ep++;
			}
			if (do_print) {
			    printf("ENDP, addr 0x%x %s pkt 0x%x int %d",
				ed->bEndpointAddress,
				eptype(ed->bmAttributes),
				UGETW(ed->wMaxPacketSize),
				ed->bInterval);
			}
			break;
		}
		if (do_print)
		    printf("\n");
        }
	if (do_print)
		printf("-- total %d interfaces, if %d: %d alt %d endp\n",
			nintf, ifno, nalt, nep);
	if (pass == 0 && ld != NULL) {
		/* allocate memory. The first part is for nalt interface
		 * settings, the second part is for the total number of
		 * endpoints. neps[i] has the total for each altsetting.
		 */
		int sz = nalt * sizeof(struct usb_host_interface) +
			nep * sizeof(struct usb_host_endpoint);

		ld->mem = malloc(sz, M_USBDEV, M_WAITOK | M_ZERO);
		printf("--- allocate %d bytes gives %p\n", sz, ld->mem);
		if (ld->mem == NULL)
			return;

		/* set ptrs to next free host_if and endpoint descriptors */
		next_alt = ld->intf.altsetting = ld->mem;
		next_ep = (struct usb_host_endpoint *)(next_alt + nalt);
		/* cur_altsetting is updated later */
		ld->intf.cur_altsetting = ld->mem;
		nalt = 0;
	}
    }
}

/*
 * Fill the linux_devinfo with information taken from the linker set.
 * XXX the linker set name must become an argument so this function
 * can be compiled as a library.
 */
void
linux_devinfo_init(struct linux_dev_info *i)
{
	struct bsd_module_init_t **p;

	if (i->initialized)
		return;
	/* initialize the struct from the linker set */
	SET_FOREACH(p, ldev_info) {
		switch ((*p)->type) {
		default:
			break;

		case BSD_MOD_LOAD_HANDLER:
			i->init_f = (*p)->value;
			break;

		case BSD_MOD_UNLOAD_HANDLER:
			i->exit_f = (*p)->value;
			break;

		case BSD_MOD_DEVICE_TABLE:
			i->device_table = (*p)->value;
			break;
		}
	}
	i->initialized = 1;
}

/*
 * Lookup a device in the table, return NULL on failure,
 * the entry on success.
 */
const struct usb_device_id *
ldev_lookup(device_t dev, struct usb_device_id *p)
{
	int i;
	struct usb_attach_arg *uaa = device_get_ivars(dev);

	parse_full_config(usbd_get_config_descriptor(uaa->device),
		uaa->ifaceno, NULL);

	if (p == NULL) {
		printf("sorry no linux device table\n");
		return NULL;
	}
	for (i = 0; p[i].match_flags != 0; i++) {
		if (p[i].match_flags & USB_DEVICE_ID_MATCH_VENDOR &&
			    p[i].idVendor != uaa->vendor)
			continue;
		if (p[i].match_flags & USB_DEVICE_ID_MATCH_PRODUCT &&
			    p[i].idProduct != uaa->product)
			continue;
		return p+i;
	}
	return NULL;
}

int
fill_linux_descriptors(device_t dev, struct linux_usb_descriptors *ld)
{
	struct usb_attach_arg *uaa = device_get_ivars(dev);
	usbd_device_handle udev = uaa->device;
	usb_config_descriptor_t *cdesc = usbd_get_config_descriptor(udev);

	bzero(ld, sizeof(*ld));
	/* initialize FreeBSD fields */
	ld->udev = udev;
	ld->iface = uaa->iface;

	/* parent struct usb_device (the full device) */
	ld->p_dev.dev.parent = NULL;	// XXX not sure
	ld->p_dev.dev.driver_data = NULL; // filled later with linux softc
	ld->p_dev.descriptor = *(struct usb_device_descriptor *)
		usbd_get_device_descriptor(udev);
	// ld->p_dev.devnum = udev->address; // hopefully unused
	/* done */

	/* struct usb_interface fields */
	ld->intf.dev.parent = &ld->p_dev.dev;
	ld->intf.num_altsetting = usbd_get_no_alts(cdesc, uaa->ifaceno);

	parse_full_config(cdesc, uaa->ifaceno, ld);
	if (ld->mem == NULL)
		return ENOMEM;	/* failed */

	printf("interface %d has %d altsettings (cur %d)\n",
		uaa->ifaceno, ld->intf.num_altsetting,
		usbd_get_interface_altindex(uaa->iface));

	/* point to the current altsetting */
	ld->intf.cur_altsetting = ld->intf.altsetting +
		usbd_get_interface_altindex(uaa->iface);
	ld->intf.cur_altsetting->desc = *(struct usb_interface_descriptor *)
		usbd_get_interface_descriptor(uaa->iface);
	/* done */


	return 0;
}

/*
 * Here are the linux api functions that are not implemented as macros.
 */

void
usb_deregister(struct usb_driver *driver)
{
	DBG("not complete");
}

/*
 * Send a control message.
 * XXX The 'pipe' parameter is ignored, which can be a problem.
 * XXX timeout was not present in the original, but it is in ms anyways.
 */
int
usb_control_msg(struct usb_device *dev, unsigned int pipe,
        __u8 request, __u8 requesttype, __u16 value, __u16 index,
        void *data, __u16 size, int timeout)
{
        usb_device_request_t req;
	struct linux_usb_descriptors *lud;
	int ret;

	lud = container_of(dev,
		struct linux_usb_descriptors, p_dev);
        if ((pipe & 0xf) != 0)
                printf("warning, usb_control_msg for pipe 0x%x\n", pipe);
	pipe &= 0xf;	/* take only the low bits */
        req.bRequest = request;
        req.bmRequestType = requesttype;
        USETW(req.wValue, value);
        USETW(req.wIndex, index);
        USETW(req.wLength,size);
#if 0	/* if we want to specify a different pipe... */
        //return  -usbd_do_request_flags_pipe(udev,
		udev->default_pipe, &req, data, 0, NULL, timeout);
#endif
	if (0)
	    DBG("usbd_do_request_flags 0x%02x 0x%02x 0x%02x 0x%04x 0x%04x %10p %04x\n",
		pipe, request, requesttype, value, index, data, size);
	ret = usbd_do_request_flags(ludev_to_udev(dev), &req, data, 0, NULL, timeout);
	return ret;
}

/*
 * Return the descriptor for the interface
 */
struct usb_interface *
usb_ifnum_to_if(struct usb_device *dev, unsigned ifnum)
{
	/* By now assume we have only one interface... */
	return &( ludev_to_l_u_d(dev)->intf );
}

/*
 * Return a pointer to the setting for this interface.
 */
struct usb_host_interface *usb_altnum_to_altsetting(
                const struct usb_interface *intf, unsigned int altnum)
{
	return intf->altsetting + altnum;
}

int
usb_clear_halt(struct usb_device *dev, int pipe)
{
	DBG("not complete\n");
	return -EINVAL;
}

/*
 * This function sets a generic interface. This emulation layer
 * only supports one interface, so we ignore the 'ifnum' argument.
 * (XXX this needs to be fixed later).
 */
int
usb_set_interface(struct usb_device *dev, int ifnum, int alternate)
{
	int err;
	struct usb_interface *intf = &(ludev_to_l_u_d(dev)->intf);

	if (0)
	    DBG("usbd_set_interface ifnum %d to alt %d\n", ifnum, alternate);
	err = -usbd_set_interface(ludev_to_iface(dev), alternate);
	intf->cur_altsetting = intf->altsetting + alternate;
	return err;
}

/* usb buffers */
/**
 * This is the linux description:
 * usb_buffer_alloc - allocate dma-consistent buffer for URB_NO_xxx_DMA_MAP
 * @dev: device the buffer will be used with
 * @size: requested buffer size
 * @mem_flags: affect whether allocation may block
 * @dma: used to return DMA address of buffer
 *
 * Return value is either null (indicating no buffer could be allocated), or
 * the cpu-space pointer to a buffer that may be used to perform DMA to the
 * specified device.  Such cpu-space buffers are returned along with the DMA
 * address (through the pointer provided).
 *
 * These buffers are used with URB_NO_xxx_DMA_MAP set in urb->transfer_flags
 * to avoid behaviors like using "DMA bounce buffers", or tying down I/O
 * mapping hardware for long idle periods.  The implementation varies between
 * platforms, depending on details of how DMA will work to this device.
 * Using these buffers also helps prevent cacheline sharing problems on
 * architectures where CPU caches are not DMA-coherent.
 *
 * When the buffer is no longer used, free it with usb_buffer_free().
 *
 * On FreeBSD:
 *   the buffer is allocated and linked to a usbd_xfer_handle x

	x = usbd_alloc_xfer(sc->udev);
	data = usbd_alloc_buffer(x, n_isoc*max_isoc_size);

 * In the emulation layer, we use the transfer_dma field as an alias to
 * usbd_xfer_handle, which is initialized when we call usb_buffer_alloc()
 * (when we have enough information to hook into the device).
 * XXX  TODO: honor the flags for malloc ?
 */
void *
usb_buffer_alloc(struct usb_device *dev, size_t size,
	gfp_t mem_flags, dma_addr_t *dma)
{
	usbd_xfer_handle *xfer = (usbd_xfer_handle *)dma;
	usbd_device_handle udev = ludev_to_udev(dev);
	void *ret;

	if (*xfer != NULL) { /* not an error, but unusual */
		DBG("xfer already allocated %p\n", *xfer);
	} else {	/* allocate a fresh xfer descriptor */
		*xfer = usbd_alloc_xfer(udev);
		if (*xfer == NULL) {
			printf("%s: usbd_alloc_xfer failed\n", __FUNCTION__);
			return NULL;
		}
	}
	ret = usbd_alloc_buffer(*xfer, size);
	if (0)
	    DBG("usbd_alloc_buffer size %d flags 0x%x xfer %p returns %p\n",
		size, mem_flags, *xfer, ret);
	return ret;
}

void
usb_buffer_free(struct usb_device *dev, size_t size,
        void *addr, dma_addr_t dma)
{
	usbd_xfer_handle xfer = (usbd_xfer_handle )dma;

	if (0)
	    DBG("usbd_free_buffer at %p for xfer %p\n",
		usbd_get_buffer(xfer), xfer);
	usbd_free_buffer(xfer);
}


/* urb functions */

/*
 * usb_init_urb does almost nothing in our emulation. We keep it separate
 * in case someone calls it directly (then we must make it global).
 */
static void
usb_init_urb(struct urb *urb)
{
	bzero(urb, sizeof (*urb));
}

/*
 * Allocate a control block and descriptors for 'iso_packets'
 * packets to be used in the transfer.
 * When we complete kmalloc we also honor the mem_flags.
 */
struct urb *
usb_alloc_urb(int iso_packets, gfp_t mem_flags)
{
	struct urb *urb = malloc(sizeof(struct urb) +
                iso_packets * sizeof(struct usb_iso_packet_descriptor),
		M_USBDEV, M_WAITOK | M_ZERO);
        if (!urb) {
                DBG("malloc failed\n");
                return NULL;
        }
        usb_init_urb(urb);
        return urb;
}

/*
 * Free the urb and the fields (xfer, frlengths[]) possibly allocated in it.
 */
void
usb_free_urb(struct urb *urb)
{
	usbd_xfer_handle xfer = (usbd_xfer_handle )urb->transfer_dma;
	if (xfer) {
		if (0)
		    DBG("usbd_free_xfer %p\n", xfer);
		usbd_free_xfer(xfer);
		urb->transfer_dma = 0 /* is this an int ??? */;
	}
	if (urb->frlengths != NULL)
		free(urb->frlengths, M_USBDEV);
	free(urb, M_USBDEV);
}

static void
ldev_isoc_handler(usbd_xfer_handle xfer, usbd_private_handle addr,
	usbd_status status)
{
	struct urb *urb = (struct urb *)addr;
	int totlen = 0, i;
	void *bsd_buf = usbd_get_buffer(xfer);

	if (status == USBD_CANCELLED) {
		DBG("usbd_cancelled %p\n", xfer);
		return;
	}
	if (status != USBD_NORMAL_COMPLETION) {
		DBG("status %d\n", status);
		return;
	}
	if (bsd_buf != urb->transfer_buffer) {
		if (0)
		    DBG("bsd != linux buffer (%p != %p)\n",
			bsd_buf, urb->transfer_buffer);
		/* XXX check direction. this is good for pipe reads */
		bcopy(bsd_buf, urb->transfer_buffer,
			urb->transfer_buffer_length);
	}
	/* map status and sizes as needed */
	for (i=0; i < urb->number_of_packets; i++) {
		totlen += urb->frlengths[i];
		urb->iso_frame_desc[i].actual_length = urb->frlengths[i];
		urb->iso_frame_desc[i].status = 0; /* XXX fake ? */
	}
	// usbd_dump_pipe(ludev_to_l_u_d(urb->dev)->bsd_pipe); // XXX debugging
	if (0)
	    DBG("status %d bytes %d for urb %p\n",
		status, totlen, urb);
	urb->status = status; /* XXX remap ? */
	urb->complete(urb, NULL /* pt_regs */);
}

/*
 * From the linux documentation:
 * usb_submit_urb - issue an asynchronous transfer request for an endpoint
 * @urb: pointer to the urb describing the request
 * @mem_flags: the type of memory to allocate, see kmalloc() for a list
 *      of valid options for this.
 *
 * This submits a transfer request, and transfers control of the URB
 * describing that request to the USB subsystem.  Request completion will
 * be indicated later, asynchronously, by calling the completion handler.
 * The three types of completion are success, error, and unlink
 * (a software-induced fault, also called "request cancellation").
 * ...
 * On Linux:
 *	this is called the first time in the open() or equivalent
 *	routine, after usb_alloc_urb(), usb_buffer_alloc(), and
 *	after the application has manually set the fields of the urb
 *	(dev, context, pipe, transfer_flags, interval, complete,
 *	number_of_packets, transfer_buffer_length, and
 *	iso_frame_desc[i]{.offset, .length}
 *
 *	In the next rounds, (typically from the 'complete' routine)
 *	the caller sets urb->dev (why ?) and clears urb->status,
 *	and reinvokes the function.
 *
 * On FreeBSD:
 *	The manual setting of fields is replaced by
 *	usbd_setup_XXX_xfer(), then this function is equivalent to
 *	usbd_transfer()
 *	XXX is the type of transfer (bulk, control, intr, isoc).
 *	Note, on FreeBSD the individual descriptors are in a private array,
 *	so we need to allocate them here.
 *
 *	Note 2 - if somebody allocates the buffer externally, i.e. without
 *	usb_buffer_alloc(), we allocate a bounce buffer to get the
 *	necessary DMA pointers. There might be better ways.
 */
int
usb_submit_urb(struct urb *urb, gfp_t mem_flags)
{
	int i, totlen;
	int xfertype = -1;	/* invalid */
	usbd_xfer_handle xfer = (usbd_xfer_handle )urb->transfer_dma;
	usbd_pipe_handle *bsd_pipe = &(ludev_to_l_u_d(urb->dev)->bsd_pipe);
	void *bsd_buf;

	if (xfer == NULL) {
		DBG("urb buffer not allocated, error\n");
		usb_buffer_alloc(urb->dev, urb->transfer_buffer_length,
			GFP_KERNEL, &urb->transfer_dma);
		xfer = (usbd_xfer_handle )urb->transfer_dma;
	}
	bsd_buf = usbd_get_buffer(xfer);

	if (bsd_buf == NULL) {
		printf("%s: transfer buffer not allocated\n", __FUNCTION__);
		return -EINVAL;
	}
	if (bsd_buf != urb->transfer_buffer) {
		DBG("bsd != linux buffer (%p != %p)\n",
			bsd_buf, urb->transfer_buffer);
	}
	if (*bsd_pipe == NULL) {	/* open the pipe */
		usbd_interface_handle iface = ludev_to_iface(urb->dev);
		int err = usbd_open_pipe(iface, urb->pipe, 0, bsd_pipe);
		if (0)
		    DBG("usbd_open_pipe 0x%x speed %d attributes %d\n",
			urb->pipe,
			usbd_get_speed(ludev_to_udev(urb->dev)),
			usbd_get_endpoint_descriptor(iface,
				urb->pipe)->bmAttributes
			);
		if (err) {
			DBG("openpipe error %d\n", err);
			return -err;
		}
	}
	if (urb->number_of_packets > 0)
		xfertype = UE_ISOCHRONOUS;
	switch (xfertype) {
	default:
		DBG("xfertype %d not supported\n", xfertype);
		return -EINVAL;

	/* XXX todo: handlers for other types */
	case UE_ISOCHRONOUS:
		/* copy buffer sizes etc */
		if (urb->number_of_packets > urb->n_frames) {
			if (urb->frlengths != NULL)
				free(urb->frlengths, M_USBDEV);
			urb->frlengths = malloc(sizeof(uint16_t) *
				urb->number_of_packets, M_USBDEV, M_NOWAIT);
		}
		if (urb->frlengths == NULL) {
			DBG("cannot allocate pointers\n");
			return -EINVAL;
		}
		urb->n_frames = urb->number_of_packets;
		totlen = 0;
		for (i=0; i < urb->number_of_packets; i++) {
			/* XXX probably something is redundant */
			urb->frlengths[i] = urb->iso_frame_desc[i].length;
			urb->iso_frame_desc[i].offset = totlen;
			urb->iso_frame_desc[i].actual_length = 0;
			urb->iso_frame_desc[i].status = -EXDEV;
			totlen += urb->frlengths[i];
		}
		if (0)
		    DBG("usbd_setup_isoc_xfer %p\n", xfer);
		usbd_setup_isoc_xfer(xfer, *bsd_pipe,
			urb, urb->frlengths, urb->number_of_packets,
			USBD_NO_COPY, ldev_isoc_handler);
		if (0)
		    DBG("usbd_transfer tot %d %p\n", totlen, xfer);
		usbd_transfer(xfer); /* error return ? */
		return 0;
	}
	DBG("not complete\n");
	return -EINVAL;
}

/**
 * usb_unlink_urb - abort/cancel a transfer request for an endpoint  
 * @urb: pointer to urb describing a previously submitted request,
 *      may be NULL
 *
 * This routine cancels an in-progress request.  URBs complete only   
 * once per submission, and may be canceled only once per submission.
 */
int
usb_unlink_urb(struct urb *urb)
{
	usbd_pipe_handle *bsd_pipe = &(ludev_to_l_u_d(urb->dev)->bsd_pipe);

        if (*bsd_pipe == NULL) {	/* pipe already closed */
		return -EINVAL;	/* or maybe just 0 ? */
	}
	if (0)
	    DBG("usbd_abort_pipe\n");
	usbd_abort_pipe(*bsd_pipe);
	if (0)
	    DBG("usbd_close_pipe\n");
	usbd_close_pipe(*bsd_pipe);
	*bsd_pipe = NULL;
	return 0;
}

/**
 * usb_kill_urb - cancel a transfer request and wait for it to finish
 * @urb: pointer to URB describing a previously submitted request,
 *      may be NULL
 */

void
usb_kill_urb(struct urb *urb)
{
	DBG("not complete but try something\n");
	usb_unlink_urb(urb);
}



/* vmalloc/vfree */
/*
 * vmalloc_32() allocates memory, 32-bit addressable.
 * On linux it is simply __vmalloc(size, GFP_KERNEL, PAGE_KERNEL);
 * For the time being, just map it to malloc()
 */
void *
vmalloc_32(unsigned long size)
{
	return malloc(size, M_USBDEV, M_WAITOK | M_ZERO);
}

void vfree(void *ptr)
{
	free(ptr, M_USBDEV);
}

/*
 * Map a vmalloc()-space virtual address to the physical page.
 */
struct page *
vmalloc_to_page(void *addr)
{
	DBG("not complete\n");
	return NULL;
}

void *
page_address(struct page *page)
{
	DBG("not complete\n");
	return NULL;
}

int
remap_pfn_range(struct vm_area_struct *vma, unsigned long addr,
                        unsigned long pfn, unsigned long size, pgprot_t pgr)
{
	DBG("not complete\n");
	return -EINVAL;
}

/* jiffies */
u64
get_jiffies_64(void)
{
	return ticks;
}

int jiffies_to_msecs(const u64 j)
{
	return ticks * 1000 / hz;
}

/* tasklet */

/*
 * On linux this is probably meant to be scheduled independently of the
 * main task. But for the time being, let's just do it as a direct call.
 */
void
tasklet_schedule(struct tasklet_struct *t)
{
	t->func(t->data);
}

void tasklet_init(struct tasklet_struct *t,
                         void (*func)(unsigned long), unsigned long data)
{
	t->func = func;
	t->data = data;
}

void
schedule(void)
{
	DBG("not complete\n");
}

/* video, mostly from videodev.c */
/*
 * on linux, we have the video_device[] array which is set
 * in video_register_device() and read with video_devdata()
 */
#define VIDEO_NUM_DEVICES       256
#define VIDEO_NAME              "video4linux"

static struct video_device *video_device[VIDEO_NUM_DEVICES];
// static DEFINE_MUTEX(videodev_lock);

struct video_device *
video_devdata(struct file *f)
{
	int i = 0;
	DBG("not complete but ok for now\n");
	return video_device[i];
}

/*
 * On linux, this does the copyin/copyout around the invocation of
 * the actual invocation. On FreeBSD, we don't need to copy.
 * Note that 'inode' is the softc.
 */
int
video_usercopy(struct inode *inode, struct file *file,
                          unsigned int cmd, unsigned long arg,
                          int (*func)(struct inode *inode, struct file *file,
                                      unsigned int cmd, void *arg))
{
	return func(inode, file, cmd, (void *)arg);
}

/*
 * register a video device with the system.
 * There is some complex linkage here, but ultimately:
 * - d->fops is the struct file_operations;
 *
 * - d->dev points to the parent struct device, in turn
 *	contained into a struct usb_device, in turn (on FreeBSD)
 *	contained in a struct linux_usb_descriptors, in turn
 *	contained in the softc, so we can write d->fops in the softc.
 */
int
video_register_device(struct video_device *d, int type, int nr)
{
	struct usb_device *udev = container_of(d->dev, struct usb_device, dev);
	struct linux_usb_descriptors *l_u_d = ludev_to_l_u_d(udev);

	l_u_d->fops = d->fops;
	video_device[0] = d;	/* XXX to be fixed */
	DBG("to be fixed but ok for now\n");
	
	return 0;
}

void
video_unregister_device(struct video_device *d)
{
	video_device[0] = NULL;
	DBG("to be fixed but ok for now\n");
}

/*
 * Allocate enough memory to store the video device data.
 * Note that this is filled in with the template (including fops)
 * in the _probe routine (i.e. attach, in FreeBSD terminology)
 */
struct video_device *
video_device_alloc(void)
{
	return malloc(sizeof(struct video_device), M_USBDEV, M_WAITOK | M_ZERO);
}

void
video_device_release(struct video_device *vfd)
{
	free(vfd, M_USBDEV);
}

/* in uaccess.h */
/*
 * copy a buffer to userspace.
 * In our emulation, *to is really a uio argument.
 * XXX TODO: check direction of transfer; return # of bytes not copied.
 */
unsigned long
copy_to_user(void __user *to, const void *from, unsigned long n)
{
	struct uio *uio = to;
	int i, prev = uio->uio_resid;
	int err;
	void *src = (void *)(uintptr_t)from;

	/* XXX should check the direction. we trust it is out or UIO_READ*/
	err = uiomove(src, n, uio);
	i = (prev - uio->uio_resid);
	if (i != n)
		DBG("uiomove asked %ld bytes, uio has %d, done %d\n",
			n, uio->uio_resid, i);
	return n - i;
}

unsigned long
copy_from_user(void *to, const void __user *from, unsigned long n)
{
	DBG("not complete\n");
	return -EINVAL;
}
